#include "exp_board_config.h"
#include "adc.h"
#include "mem.h"
#include "ports.h"

// number of channels to convert
#define NUM_CH 8
// number of samples to average
#ifndef ADC_MAX_NUM_SAMPLES
  #define ADC_MAX_NUM_SAMPLES                  16
#endif

// private variables
volatile uint8_t adc_sample_period_ms;
// current channel
volatile uint8_t adc_current_ch;
// channel data
volatile int16_t adc_data[NUM_CH];
// averaged channel data
volatile int16_t adc_ch_data[NUM_CH][ADC_MAX_NUM_SAMPLES];
volatile uint8_t adc_current_sample;
volatile uint8_t adc_num_samples;
volatile uint8_t adc_running;
volatile uint8_t adc_prescaler_mask;

/* private functions */
void adc_set_channel(uint8_t ch_id)
{
  ADMUX&=0xF8;
  ADMUX|= (ch_id&0x07);
}

void adc_start_conv(void)
{
  ADCSRA|=(1<<ADSC);
}

uint8_t adc_is_conversion_done(void)
{
  if(ADCSRA&(1<<ADIF))
    return 0x01;
  else
    return 0x00;
}

uint8_t adc_is_period_done(void)
{
  if(TIFR&(1<<OCF2))
    return 0x01;
  else
    return 0x00;
}

/* public functions */
void adc_init(void)
{
  // configure the analog inputs
  DDRF = 0x00;
  PORTF = 0x00;  

  ADMUX = (1<<REFS0);// use AVCC as reference, Right justified, channel 0
  ADCSRA = (1<<ADPS2) | (1<<ADPS1) | (1<<ADPS0); // free running mode disabled, interrupts enabled, prescaler to 128

  // clear any pending interrupt
  ADCSRA|=(1<<ADIF);

  // set the initial channel to convert
  adc_current_ch=0;
  adc_current_sample=0;
  adc_running=0;
  adc_num_samples=ADC_MAX_NUM_SAMPLES;
  ram_data[ADC_NUM_SAMPLES]=ADC_MAX_NUM_SAMPLES;
  adc_sample_period_ms=100;
  ram_data[ADC_SAMPLE_PERIOD]=100;
  adc_set_sample_period(adc_sample_period_ms);

  // configure the timer 2 to perform periodic conversions (1 ms period)
  TCCR2=(1<<WGM21);// CTC mode, no output, prescaler to 0 (no clock source)
  TCNT2=0x00;// start from 0
}

void adc_start(void)
{
  // enable the ADC 
  ADCSRA|=(1<<ADEN);
  // enable the timer 2 interrupts
  TCNT2=0x00;// start from 0
  TIMSK=0x00;// interrupts not enabled
  // clear interrupts
  TIFR|=(1<<OCF2);
  TIFR|=(1<<TOV2);
  // configure the timer prescaler and count value
  adc_running=0x01;
  adc_set_sample_period(adc_sample_period_ms);
  ram_data[ADC_CONTROL]|=0x04;
}

void adc_stop(void)
{
  TCCR2&=0xF8;// set prescaler to 0 (no clock source)
  // disable the ADC
  ADCSRA&=~(1<<ADEN);
  adc_running=0x00;
  ram_data[ADC_CONTROL]&=0xFB;
}

void adc_set_num_samples(uint8_t num)
{
  if(num>ADC_MAX_NUM_SAMPLES)
    adc_num_samples=ADC_MAX_NUM_SAMPLES;
  else
  {
    if(num<=0)
      adc_num_samples=1;
    else
      adc_num_samples=num;
  }
  if(adc_current_sample>num)
    adc_current_sample=0;
  ram_data[ADC_NUM_SAMPLES]=num;
}

void adc_set_sample_period(uint8_t time_ms)
{
  uint16_t prescaler_values[5]={1,8,64,256,1024};
  uint16_t compare_value;
  uint8_t i;

  for(i=0;i<5;i++)
  {
    compare_value=(((F_CPU/1000)*time_ms)/(prescaler_values[i]*NUM_CH))-1;
    if(compare_value<256)
    {
      OCR2=(uint8_t)compare_value;
      if(adc_running)
      {
        TCCR2&=0xF8;
        TCCR2|=(i+1);
      }
      else
        adc_prescaler_mask=i+1;
      break;
    }
  }
  ram_data[ADC_SAMPLE_PERIOD]=time_ms;
}

void adc_loop(void)
{
  uint8_t i;
  uint16_t data;

  if(adc_is_period_done())
  {
    TIFR|=(1<<OCF2);// clear interrupt flag
    TIFR|=(1<<TOV2);
    TCNT2=0x00;// start from 0
    // start a new conversion
    adc_set_channel(adc_current_ch);
    adc_start_conv();
  }
  else
  {
    if(adc_is_conversion_done())
    {
      ADCSRA|=(1<<ADIF);// clear interrupt flag
      data=(ADCL | (ADCH<<8));
      ram_data[ADC_CHANNEL0_L+adc_current_ch*2]=data%256;
      ram_data[ADC_CHANNEL0_H+adc_current_ch*2]=data/256;
      ram_data[ADC_CHANNEL0_L_block+adc_current_ch*2]=data%256;
      ram_data[ADC_CHANNEL0_H_block+adc_current_ch*2]=data/256;
      adc_ch_data[adc_current_ch][adc_current_sample]=data;
      // compute the average for the current channel
      data=0;
      for(i=0;i<adc_num_samples;i++)
        data+=adc_ch_data[adc_current_ch][i];
      data=data/adc_num_samples;
      ram_data[ADC_AVG_CHANNEL0_L+adc_current_ch*2]=data%256;
      ram_data[ADC_AVG_CHANNEL0_H+adc_current_ch*2]=data/256;
      ram_data[ADC_AVG_CHANNEL0_L_block+adc_current_ch*2]=data%256;
      ram_data[ADC_AVG_CHANNEL0_H_block+adc_current_ch*2]=data/256;
      if(adc_current_ch==NUM_CH-1)// last channel
        adc_current_sample=(adc_current_sample+1)%adc_num_samples;
      adc_current_ch=(adc_current_ch+1)%NUM_CH;
    }
  }
}
