#include "exp_board_config.h"
#include "bno055.h"
#include "i2c.h"
#include "mem.h"
#include "ports.h"
#include <util/delay.h>
#include <avr/eeprom.h>

typedef enum {get_calibrated,is_calibrated,idle,get_data,read_data} bno055_states_t;

/* private variables */
volatile uint8_t bno055_stop_op;
volatile uint8_t bno055_start_op;
volatile uint8_t bno055_detected;
volatile bno055_states_t bno055_state;

uint8_t EEMEM bno055_cal_data[24]={0xFF};
//uint8_t EEMEM bno055_cal_data[24]={0x55,0xAA,// eeprom header to detect valid data
//                                   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xFE,0xFF,0x00,0x00,0xE8,0x03,0x45,0x02,0x0A};

/* private functions */
void bno055_convert_quaternion_data(uint8_t *data)
{
  ram_data[IMU_QUAT_X_L]=data[2];
  ram_data[IMU_QUAT_X_H]=data[3];
  ram_data[IMU_QUAT_Y_L]=data[4];
  ram_data[IMU_QUAT_Y_H]=data[5];
  ram_data[IMU_QUAT_Z_L]=data[6];
  ram_data[IMU_QUAT_Z_H]=data[7];
  ram_data[IMU_QUAT_W_L]=data[0];
  ram_data[IMU_QUAT_W_H]=data[1];
}

void bno055_convert_euler_angles_data(unsigned char *data)
{
  short int angle;

  angle=((short int)(data[0]+(data[1]<<8))*10)/16;
  ram_data[IMU_YAW_L]=angle%256;
  ram_data[IMU_YAW_H]=angle/256;
  angle=((short int)(data[4]+(data[5]<<8))*10)/16;
  ram_data[IMU_PITCH_L]=angle%256;
  ram_data[IMU_PITCH_H]=angle/256;
  angle=((short int)(data[2]+(data[3]<<8))*10)/16;
  ram_data[IMU_ROLL_L]=angle%256;
  ram_data[IMU_ROLL_H]=angle/256;
}

void bno055_convert_linear_accel_data(unsigned char *data)
{
  short int acc;

  acc=((short int)(data[0]+(data[1]<<8))*10);
  ram_data[IMU_ACCEL_X_L]=acc%256;
  ram_data[IMU_ACCEL_X_H]=acc/256;
  acc=((short int)(data[2]+(data[3]<<8))*10);
  ram_data[IMU_ACCEL_Y_L]=acc%256;
  ram_data[IMU_ACCEL_Y_H]=acc/256;
  acc=((short int)(data[4]+(data[5]<<8))*10);
  ram_data[IMU_ACCEL_Z_L]=acc%256;
  ram_data[IMU_ACCEL_Z_H]=acc/256;
}

void bno055_convert_gravity_data(unsigned char *data)
{
  short int grav;

  grav=((short int)(data[0]+(data[1]<<8))*10);
  ram_data[IMU_GRAV_X_L]=grav%256;
  ram_data[IMU_GRAV_X_H]=grav/256;
  grav=((short int)(data[2]+(data[3]<<8))*10);
  ram_data[IMU_GRAV_Y_L]=grav%256;
  ram_data[IMU_GRAV_Y_H]=grav/256;
  grav=((short int)(data[4]+(data[5]<<8))*10);
  ram_data[IMU_GRAV_Z_L]=grav%256;
  ram_data[IMU_GRAV_Z_H]=grav/256;
}


void bno055_load_calibration(void)
{
  uint8_t cmd[23],i;
  uint8_t byte1,byte2;
 
  // read the eeprom for a valid calibration
  byte1=eeprom_read_byte(&bno055_cal_data[0]);
  byte2=eeprom_read_byte(&bno055_cal_data[1]);
  if(byte1==0x55 && byte2==0xAA)// a valid configuration is in memory
  {
    cmd[0]=0x55;
    for(i=0;i<22;i++)
      cmd[i+1]=eeprom_read_byte(&bno055_cal_data[i+2]);
    i2c_master_send(IMU_ADDR,23,cmd);
    while(i2c_get_state()!=I2C_IDLE);
  }
}

void bno055_save_calibration(void)
{
  uint8_t i,cal_data[22],cmd=0x55;

  /* write the valid calibration header */
  eeprom_write_byte(&bno055_cal_data[0],0x55);
  eeprom_write_byte(&bno055_cal_data[1],0xAA);
  /* read the calibration data from the IMU */
  i2c_master_reg_receive(IMU_ADDR,cmd,22);
  while(i2c_get_state()!=I2C_IDLE);
  i2c_get_data(22,cal_data);
  /* write the calibration data */
  for(i=0;i<22;i++)
    eeprom_write_byte(&bno055_cal_data[i+2],cal_data[i]);
}

void bno055_erase_calibration(void)
{
  uint8_t i;

  for(i=0;i<2;i++)
    eeprom_write_byte(&bno055_cal_data[i],0xFF); 
}

void bno055_change_register_page(unsigned char page_id)
{
  uint8_t cmd[2];
 
  cmd[0]=0x07;
  cmd[1]=page_id;
  i2c_master_send(IMU_ADDR,2,cmd);
  while(i2c_get_state()!=I2C_IDLE);
}

void bno055_set_mode(op_mode_t mode)
{
  uint8_t cmd[2];
 
  cmd[0]=0x3D;
  cmd[1]=(unsigned char)mode;
  i2c_master_send(IMU_ADDR,2,cmd);
  while(i2c_get_state()!=I2C_IDLE);
}

void bno055_set_power_mode(power_mode_t mode)
{
  uint8_t cmd[2];
 
  cmd[0]=0x3E;
  cmd[1]=(unsigned char)mode;
  i2c_master_send(IMU_ADDR,2,cmd);
  while(i2c_get_state()!=I2C_IDLE);
}

uint8_t bno055_detect(void)
{
  uint8_t cmd=0x01,info[6];

  i2c_master_reg_receive(IMU_ADDR,cmd,6);
  while(i2c_get_state()!=I2C_IDLE);
  i2c_get_data(6,info);
  if(info[0]==0xFB && info[1]==0x32 && info[2]==0x0F)
    return 0x01;
  else
    return 0x00;
}

/* public functions */
void bno055_init(void)
{
  /* initialize private variables */
  bno055_stop_op=0;
  bno055_start_op=0;
  bno055_state=get_calibrated;
  bno055_detected=0x00;

  /* initialize i2c peripheral */
  i2c_init(100);

  /* configure IMU */
  if(bno055_detect())
  {
    // configure the device 
    bno055_set_mode(config_mode);
    bno055_change_register_page(0); 
    bno055_load_calibration();
    // start the device
    bno055_set_mode(ndof_mode);
    bno055_detected=0x01;
    ram_data[IMU_CONTROL]|=0x80;
  }
  else
  {
    bno055_detected=0x00;
    ram_data[IMU_CONTROL]&=0x7F;
  }
}

uint8_t bno055_is_detected(void)
{
  return bno055_detected;
}

void bno055_start(void)
{
  bno055_stop_op=0;
  bno055_start_op=1;
  ram_data[IMU_CONTROL]|=0x08;
}

void bno055_stop(void)
{
  bno055_stop_op=1;
  bno055_start_op=0;
}

void bno055_loop(void)
{
  uint8_t cmd,imu_data[26],cal_data;

  switch(bno055_state)
  {
    case get_calibrated: cmd=0x35;
                         i2c_master_reg_receive(IMU_ADDR,cmd,1);
                         bno055_state=is_calibrated;
                         break;
    case is_calibrated: if(i2c_get_state()==I2C_IDLE)
                        {
                          i2c_get_data(1,&cal_data);
                          if((cal_data&0x30)==0x30)
                            ram_data[IMU_CONTROL]|=0x20;
                          else
                            ram_data[IMU_CONTROL]&=0xDF;
                          if((cal_data&0x0C)==0x0C)
                            ram_data[IMU_CONTROL]|=0x40;
                          else
                            ram_data[IMU_CONTROL]&=0xBF;
                          if((cal_data&0x03)==0x03)
                            ram_data[IMU_CONTROL]|=0x10;
                          else
                            ram_data[IMU_CONTROL]&=0xEF;
                          if((ram_data[IMU_CONTROL]&0x70)==0x70)
                          {
                            bno055_state=idle;
                            bno055_save_calibration();
                          }
                          else
                            bno055_state=get_calibrated;
                        }
                        else
                          bno055_state=is_calibrated;
                        break;
    case idle: if(bno055_start_op==1)
               {
                 bno055_state=get_data;
                 bno055_start_op=0;
               }
               else
                 bno055_state=idle;
               break;
    case get_data: if(i2c_get_state()==I2C_IDLE)
                   {
                     cmd=0x1A;
                     i2c_master_reg_receive(IMU_ADDR,cmd,26);
                     bno055_state=read_data;
                   }
                   else
                     bno055_state=get_data;
                   break;
    case read_data: if(i2c_get_state()==I2C_IDLE)
                    {
                      i2c_get_data(26,imu_data);
                      bno055_convert_euler_angles_data(imu_data);
                      bno055_convert_quaternion_data(&imu_data[6]);
                      bno055_convert_linear_accel_data(&imu_data[14]);
                      bno055_convert_gravity_data(&imu_data[20]);
                      if(bno055_stop_op==1)
                      {
                        bno055_stop_op=0;
                        bno055_state=idle;
                        ram_data[IMU_CONTROL]&=0xF7;
                      }
                      else
                        bno055_state=get_data;
                    }
                    else
                      bno055_state=read_data;
                    break;
  }
}
