#include "exp_board_config.h"
#include "compass.h"
#include "i2c.h"
#include "mem.h"
#include "ports.h"

#ifndef COMPASS_MAX_NUM_SAMPLES
  #define COMPASS_MAX_NUM_SAMPLES             16
#endif

typedef enum {idle,get_data,get_data_delay,read_data,start_cal,calibrating,calibration_delay,stop_cal} compass_states_t;

/* private variables */
volatile uint8_t compass_current_sample;
volatile uint8_t compass_num_samples;

volatile uint8_t compass_start_cal;
volatile uint8_t compass_stop_cal;
volatile uint8_t compass_stop_op;
volatile uint8_t compass_start_op;
volatile compass_states_t compass_state;

/* private functions */
void compass_avrg(uint16_t heading)
{
  static unsigned long int new_average,first_value;
  unsigned long int diff=0;

  if(compass_current_sample==0)
  {
    compass_current_sample++;
    first_value=(long int)heading;
  }  
  else
  {
    diff=heading+3600-first_value;
    if(diff<1800)
      diff=diff+3600;
    else if(diff>5400) 
      diff=diff-3600;
    new_average=new_average+diff;
    compass_current_sample++;
    if(compass_current_sample==compass_num_samples)
    {
      new_average=new_average/(compass_num_samples-1);
      new_average=new_average+first_value;
      if(new_average>3600)
	new_average-=3600;
      ram_data[COMPASS_AVG_HEADING_VAL_L]=new_average%256;
      ram_data[COMPASS_AVG_HEADING_VAL_H]=new_average/256;
      ram_data[COMPASS_AVG_HEADING_VAL_L_block]=new_average%256;
      ram_data[COMPASS_AVG_HEADING_VAL_H_block]=new_average/256;
      new_average=0;
      compass_current_sample=0; 
    }
  }
}

void compass_start_time_delay(uint8_t delay_ms)
{
  TCNT0=0x00;// reset counter
  OCR0=delay_ms*16;// set count value
  TCCR0&=0xF8;
  TCCR0|=0x07;// enable clock
}

uint8_t compass_is_delay_done(void)
{
  if(TIFR&(1<<OCF0))
  {
    // disable clock
    TIFR|=(1<<OCF0);
    TCCR0&=0xF8;
    return 0x01;
  }
  else
    return 0x00;
}

/* public functions */
void compass_init(void)
{
  /* initialize private variables */
  compass_start_cal=0;
  compass_stop_cal=0;
  compass_stop_op=0;
  compass_start_op=0;
  compass_state=idle;
  compass_current_sample=0;
  compass_num_samples=COMPASS_MAX_NUM_SAMPLES;
  ram_data[COMPASS_NUM_SAMPLES]=COMPASS_MAX_NUM_SAMPLES;

  /* use timer 0 to control the operation delay */
  TCCR0=(1<<WGM01);//campare mode, output disable, prescaler 0
  ASSR=0x00;// disable assynchronous operation
  // interrupts not enabled
  OCR0=0xFF;
  TCNT0=0x00;

  /* initialize i2c peripheral */
  i2c_init(100);
}

void compass_start(void)
{
  compass_stop_op=0;
  compass_start_op=1;
  ram_data[COMPASS_CONTROL]|=0x40;
}

void compass_stop(void)
{
  compass_stop_op=1;
  compass_start_op=0;
}

void compass_start_calibration(void)
{
  compass_start_cal=1;
  compass_stop_cal=0;
}

void compass_stop_calibration(void)
{
  compass_start_cal=0;
  compass_stop_cal=1;
}

void compass_set_num_samples(uint8_t num)
{
  if(num>COMPASS_MAX_NUM_SAMPLES)
    compass_num_samples=COMPASS_MAX_NUM_SAMPLES;
  else if(num<=2)
    compass_num_samples=2;
  else
    compass_num_samples=num;
  ram_data[COMPASS_NUM_SAMPLES]=compass_num_samples;
}

void compass_loop(void)
{
  uint8_t cmd,heading[2];
  unsigned short int compass_heading;
  static compass_states_t prev_state=idle;

  switch(compass_state)
  {
    case idle: if(compass_start_op==1)
               {
                 compass_state=get_data;
                 compass_start_op=0;
               }
               else
               {
                 if(compass_start_cal==1)
                 {
                   compass_start_cal=0;
                   compass_state=start_cal;
                   prev_state=idle;
                 }
                 else
                   compass_state=idle;
               }
               break;
    case get_data: if(i2c_get_state()==I2C_IDLE)
                   {
                     cmd=COMPASS_HEADING;
                     i2c_master_send(COMPASS_ADDR,1,&cmd);
                     compass_start_time_delay(10);
                     compass_state=get_data_delay;
                   }
                   else
                     compass_state=get_data;
                   break;
    case get_data_delay: if(compass_is_delay_done()==1)
                         {
                           if(i2c_get_state()==I2C_IDLE)
                           {
                             i2c_master_receive(COMPASS_ADDR,2);
                             compass_state=read_data;
                           }
                           else
                             compass_state=get_data_delay;
                         }
                         else
                           compass_state=get_data_delay;
                         break;
    case read_data: if(i2c_get_state()==I2C_IDLE)
                    {
                      i2c_get_data(2,heading);
                      compass_heading=heading[1]+(heading[0]<<8);
                      ram_data[COMPASS_HEADING_VAL_L]=heading[1];
                      ram_data[COMPASS_HEADING_VAL_H]=heading[0];
                      ram_data[COMPASS_HEADING_VAL_L_block]=heading[1];
                      ram_data[COMPASS_HEADING_VAL_H_block]=heading[0];
                      compass_avrg(compass_heading);
                      if(compass_stop_op==1)
                      {
                        compass_stop_op=0;
                        compass_state=idle;
                        ram_data[COMPASS_CONTROL]&=0xBF;
                      }
                      else
                      {
                        if(compass_start_cal==1)
                        {
                          compass_start_cal=0;
                          compass_state=start_cal;
                          prev_state=get_data;
                        }
                        else
                          compass_state=get_data;
                      }
                    }
                    else
                      compass_state=read_data;
                    break;
    case start_cal: if(i2c_get_state()==I2C_IDLE)
                    {
                      cmd=COMPASS_START_CAL;
                      i2c_master_send(COMPASS_ADDR,1,&cmd);
                      ram_data[COMPASS_CONTROL]|=0x80;
                      compass_state=calibrating;
                    }
                    else
                      compass_state=start_cal; 
                    break;
    case calibrating: if(compass_stop_cal==1)
                      {
                        compass_state=stop_cal;
                        compass_stop_cal=0;
                      }
                      else
                        compass_state=calibrating;
                      break;
    case stop_cal: if(i2c_get_state()==I2C_IDLE)
                   {
                     cmd=COMPASS_STOP_CAL;
                     i2c_master_send(COMPASS_ADDR,1,&cmd);
                     compass_start_time_delay(20);
                     compass_state=calibration_delay;
                   }
                   else
                     compass_state=stop_cal;
                   break;
    case calibration_delay: if(compass_is_delay_done()==1)
                            {
                              compass_state=prev_state;
                              ram_data[COMPASS_CONTROL]&=0x7F;
                            }
                            else
                              compass_state=calibration_delay;
                            break;
  }
}
