#include "exp_board_config.h"
#include <avr/io.h>
#include <avr/interrupt.h>
#include "i2c.h"
#include "ports.h"

#ifndef I2C_RX_BUFFER_LEN
  #define I2C_RX_BUFFER_LEN            32
#endif

#ifndef I2C_TX_BUFFER_LEN
  #define I2C_TX_BUFFER_LEN            32
#endif

/* private variables */
volatile i2c_states_t i2c_state;
volatile uint8_t i2c_device_address;
// send/transmit buffer (outgoing data)
volatile uint8_t i2c_tx_data[I2C_RX_BUFFER_LEN];
volatile uint8_t i2c_tx_data_ptr;
volatile uint8_t i2c_tx_length;
// receive buffer (incoming data)
volatile uint8_t i2c_rx_data[I2C_TX_BUFFER_LEN];
volatile uint8_t i2c_rx_data_ptr;
volatile uint8_t i2c_rx_length;

/* private functions */
inline void i2c_send_start(void)
{
  // send start condition
  TWCR=(TWCR&TWCR_CMD_MASK)|(1<<TWINT)|(1<<TWSTA)|(1<<TWEN);
}

inline void i2c_send_stop(void)
{
  // transmit stop condition
  // leave with TWEA on for slave receiving
  TWCR=(TWCR&TWCR_CMD_MASK)|(1<<TWINT)|(1<<TWEA)|(1<<TWSTO)|(1<<TWEN);
}

inline void i2c_wait_for_complete(void)
{
  // wait for i2c interface to complete operation
  while((TWCR&(1<<TWINT))==0x00);
}

inline void i2c_send_byte(uint8_t data)
{
  // save data to the TWDR
  TWDR=data;
  // begin send
  TWCR=(TWCR&TWCR_CMD_MASK)|(1<<TWINT);
}

inline void i2c_receive_byte(uint8_t ackFlag)
{
  // begin receive over i2c
  if( ackFlag )
  {
    // ackFlag = TRUE: ACK the recevied data
    TWCR=(TWCR&TWCR_CMD_MASK)|(1<<TWINT)|(1<<TWEA);
  }
  else
  {
    // ackFlag = FALSE: NACK the recevied data
    TWCR=(TWCR&TWCR_CMD_MASK)|(1<<TWINT);
  }
}

inline uint8_t i2c_get_byte(void)
{
  // retieve received data byte from i2c TWDR
  return TWDR;
}

i2c_states_t i2c_get_state(void)
{
  // retieve current i2c status from i2c TWSR
  return i2c_state;
}

/* interrupt handlers */
ISR(TWI_vect)
{
  // read status bits
  uint8_t status = TWSR&TWSR_STATUS_MASK;

  switch(status)
  {
    // Master General
    case TW_START://0x08: Sent start condition
    case TW_REP_START:// 0x10: Sent repeated start condition
      // send device address
      i2c_send_byte(i2c_device_address);
      break;
      // Master Transmitter & Receiver status codes
    case TW_MT_SLA_ACK:// 0x18: Slave address acknowledged
    case TW_MT_DATA_ACK:// 0x28: Data acknowledged
      if(i2c_tx_data_ptr < i2c_tx_length)
      {
	// send data
	i2c_send_byte( i2c_tx_data[i2c_tx_data_ptr++] );
      }
      else
      {
        if(i2c_rx_length==0)
        {
          // transmit stop condition, enable SLA ACK
          i2c_send_stop();
          // set state
          i2c_state = I2C_IDLE;
        }
        else
        {
          i2c_state = I2C_MASTER_RX;
          i2c_device_address|=0x01;       // RW set: read operation
          i2c_send_start();
        }
      }
      break;
    case TW_MR_DATA_NACK:// 0x58: Data received, NACK reply issued
      // store final received data byte
      i2c_rx_data[i2c_rx_data_ptr++] = TWDR;
      // continue to transmit STOP condition
    case TW_MR_SLA_NACK:// 0x48: Slave address not acknowledged
    case TW_MT_SLA_NACK:// 0x20: Slave address not acknowledged
    case TW_MT_DATA_NACK:// 0x30: Data not acknowledged
      // transmit stop condition, enable SLA ACK
      i2c_send_stop();
      // set state
      i2c_state = I2C_IDLE;
      break;
    case TW_MT_ARB_LOST:// 0x38: Bus arbitration lost
      // release bus
      TWCR=(TWCR&TWCR_CMD_MASK)|(1<<TWINT);
      // set state
      i2c_state = I2C_IDLE;
      break;
    case TW_MR_DATA_ACK:// 0x50: Data acknowledged
      // store received data byte
      i2c_rx_data[i2c_rx_data_ptr++] = TWDR;
      // fall-through to see if more bytes will be received
    case TW_MR_SLA_ACK:// 0x40: Slave address acknowledged
      if(i2c_rx_data_ptr < (i2c_rx_length-1))
	// data byte will be received, reply with ACK (more bytes in transfer)
	i2c_receive_byte(0x01);
      else
      {
	// data byte will be received, reply with NACK (final byte in transfer)
	i2c_receive_byte(0x00);
      }
      break;
      // Misc
    case TW_NO_INFO:// 0xF8: No relevant state information
      // do nothing
      break;
    case TW_BUS_ERROR:// 0x00: Bus error due to illegal start or stop condition
      // reset internal hardware and release bus
      TWCR=(TWCR&TWCR_CMD_MASK)|(1<<TWINT)|(1<<TWSTO)|(1<<TWEA);
      // set state
      i2c_state = I2C_IDLE;
      break;
  }
}

/* public functions */
void i2c_init(uint16_t bitrate_khz)
{
  /* configure the io ports */
  //DDRD|=0x03;
  // set i2c bit rate to 100KHz
  i2c_set_bitrate(bitrate_khz);

  /* initialize the internal variables */
  i2c_tx_data_ptr=0;
  i2c_tx_length=0;
  i2c_rx_data_ptr=0;
  i2c_rx_length=0;
  i2c_state=I2C_IDLE;

  TWCR=(1<<TWEN)|(1<<TWIE)|(1<<TWEA);
}

void i2c_set_bitrate(uint16_t bitrate_khz)
{
  uint8_t bitrate_div;

  TWSR&=~((1<<TWPS0)|(1<<TWPS1));
  // calculate bitrate division	
  bitrate_div = ((F_CPU/1000l)/bitrate_khz);
  if(bitrate_div >= 16)
    bitrate_div = (bitrate_div-16)/2;
  TWBR=bitrate_div;
}

void i2c_master_send(uint8_t deviceAddr, uint8_t length, uint8_t* data)
{
  uint8_t i;
  // wait for interface to be ready
  while(i2c_state);
  // set state
  i2c_state = I2C_MASTER_TX;
  // save data
  i2c_device_address = (deviceAddr & 0xFE);	// RW cleared: write operation
  for(i=0; i<length; i++)
    i2c_tx_data[i] = *data++;
  i2c_tx_data_ptr = 0;
  i2c_tx_length = length;
  i2c_rx_data_ptr = 0;
  i2c_rx_length = 0;
  // send start condition
  i2c_send_start();
}

void i2c_master_receive(uint8_t deviceAddr, uint8_t length)//, uint8_t* data)
{
  //	uint8_t i;
  // wait for interface to be ready
  while(i2c_state);
  // set state
  i2c_state = I2C_MASTER_RX;
  // save data
  i2c_device_address = (deviceAddr|0x01);	// RW set: read operation
  i2c_rx_data_ptr = 0;
  i2c_rx_length = length;
  i2c_tx_data_ptr = 0;
  i2c_tx_length = 0;
  // send start condition
  i2c_send_start();
}

void i2c_master_reg_receive(uint8_t deviceAddr, uint8_t reg_address, uint8_t length)//, uint8_t* data)
{
  //    uint8_t i;
  // wait for interface to be ready
  while(i2c_state);
  // set state
  i2c_state = I2C_MASTER_TX;
  // save data
  i2c_device_address = (deviceAddr & 0xFE);       // RW set: read operation
  i2c_tx_data[0]=reg_address;
  i2c_tx_data_ptr = 0;
  i2c_tx_length = 1;
  i2c_rx_data_ptr = 0;
  i2c_rx_length=length;
  // send start condition
  i2c_send_start();
}

void i2c_get_data(unsigned char length, unsigned char *data)
{
  unsigned char i;

  for(i=0;i<length;i++)
    data[i]=i2c_rx_data[i];
}

