#include "exp_board_config.h"
#include "pwm.h"
#include "mem.h"
#include "ports.h"

#ifndef PWM_DEFAULT_FREQ
  #define PWM_DEFAULT_FREQ         50// Hz
#endif

#ifndef PWM_DEFAULT_DUTY
  #define PWM_DEFAULT_DUTY         50// %
#endif

#define      PWM_MAX_CHANNELS     4

/* private variables */
uint32_t pwm_top_value;
uint8_t pwm_prescaler_mask;
uint8_t pwm_running;
uint8_t pwm_duty_cycle[PWM_MAX_CHANNELS];

/* public functions */
void pwm_init(void)
{
  /* configure the IO pins */
  DDRE|=0x38;
  DDRB|=0x80;

  /* initialize internal variables */
  pwm_top_value=40000;
  ram_data[PWM_FREQ_L]=50;
  ram_data[PWM_FREQ_H]=0;
  pwm_prescaler_mask=2;
  pwm_running=0x00;

  pwm_set_frequency(PWM_DEFAULT_FREQ);
  pwm_set_duty_cycle(pwm0,PWM_DEFAULT_DUTY);
  pwm_set_duty_cycle(pwm1,PWM_DEFAULT_DUTY);
  pwm_set_duty_cycle(pwm2,PWM_DEFAULT_DUTY);
  pwm_set_duty_cycle(pwm3,PWM_DEFAULT_DUTY);

  /* configure the timer 3 */
  TCNT3H=0x00;
  TCNT3L=0x00;
  /* set the timer mode */
  TCCR3A=(1<<WGM31);// all three channels in fast PWM mode, output disabled
  TCCR3B=(1<<WGM33)|(1<<WGM32);// no clock source, fast PWM mode
  // no interrupts enabled
}

void pwm_start(void)
{
  TCNT3H=0x00;
  TCNT3L=0x00;
  // set the prescaler value to the correct value
  /* set the prescaler value */
  TCCR3B&=0xF8;
  TCCR3B|=pwm_prescaler_mask;
  /* set the top value */
  ICR3H=(pwm_top_value>>8)&0xFF;
  ICR3L=(pwm_top_value)&0xFF;
  pwm_running=0x01;
  ram_data[PWM_CONTROL]|=0x04;
}

void pwm_stop(void)
{
  /* set the prescaler to 0 (no clock source) */
  TCCR3B&=0xF8;
  pwm_running=0x00;
  ram_data[PWM_CONTROL]&=0xFB;
}

void pwm_enable_ch(pwm_id_t id)
{
  switch(id)
  {
    case pwm0: TCCR3A|=(1<<COMA1);
               ram_data[PWM_CONTROL]|=0x10;
               break;
    case pwm1: TCCR3A|=(1<<COMB1);
               ram_data[PWM_CONTROL]|=0x20;
               break;
    case pwm2: TCCR3A|=(1<<COMC1);
               ram_data[PWM_CONTROL]|=0x40;
               break;
    case pwm3: break;
  }
}

void pwm_disable_ch(pwm_id_t id)
{
  switch(id)
  {
    case pwm0: TCCR3A&=~(1<<COMA1);
               ram_data[PWM_CONTROL]&=0xEF;
               break;
    case pwm1: TCCR3A&=~(1<<COMB1);
               ram_data[PWM_CONTROL]&=0xDF;
               break;
    case pwm2: TCCR3A&=~(1<<COMC1);
               ram_data[PWM_CONTROL]&=0xBF;
               break;
    case pwm3: break;
  }
}

void pwm_set_frequency(uint16_t freq_hz)
{
  uint32_t prescaler_values[5]={1,8,64,256,1024};
  uint32_t den;
  uint8_t i;

  for(i=0;i<5;i++)
  {
    den=freq_hz*prescaler_values[i];
    pwm_top_value=(((uint32_t)16000000)/den)-1;
    if(pwm_top_value<65536)
    {
      /* adjust the duty cyle  */
      pwm_set_duty_cycle(pwm0,pwm_duty_cycle[0]);
      pwm_set_duty_cycle(pwm1,pwm_duty_cycle[1]);
      pwm_set_duty_cycle(pwm2,pwm_duty_cycle[2]);
      pwm_set_duty_cycle(pwm3,pwm_duty_cycle[3]);
      if(pwm_running)
      {
        /* set the prescaler value */
        TCCR3B&=0xF8;
        TCCR3B|=(i+1);
        /* set the top value */
        ICR3H=(pwm_top_value>>8)&0xFF;
        ICR3L=(pwm_top_value)&0xFF;
      }
      else 
        pwm_prescaler_mask=(i+1);
      ram_data[PWM_FREQ_L]=freq_hz%256;
      ram_data[PWM_FREQ_H]=freq_hz/256;
      break;
    }
  }
}

void pwm_set_duty_cycle(pwm_id_t id,uint16_t duty)
{
  uint16_t compare_value;

  compare_value=(uint16_t)((((uint32_t)pwm_top_value*(uint32_t)duty)/100)>>9);

  switch(id)
  {
    case pwm0: OCR3AH=(compare_value>>8)&0xFF;
               OCR3AL=(compare_value&0xFF);
               pwm_duty_cycle[0]=duty;
               ram_data[PWM_CH0_DUTY]=duty;
               break;
    case pwm1: OCR3BH=(compare_value>>8)&0xFF;
               OCR3BL=(compare_value&0xFF);
               pwm_duty_cycle[1]=duty;
               ram_data[PWM_CH1_DUTY]=duty;
               break;
    case pwm2: OCR3CH=(compare_value>>8)&0xFF;
               OCR3CL=(compare_value&0xFF);
               pwm_duty_cycle[2]=duty;
               ram_data[PWM_CH2_DUTY]=duty;
               break;
    case pwm3: break;
  } 
}

