#include "exp_board_config.h"
#include <avr/io.h>
#include <avr/interrupt.h>
#include "uart_ttl.h"
#include "mem.h"

#ifndef UART_TTL_DEFAULT_BAUDRATE
  #define UART_TTL_DEFAULT_BAUDRATE          9600
#endif

#ifndef UART_TTL_RX_BUFFER_SIZE
  #define UART_TTL_RX_BUFFER_SIZE            32
#endif

#define UART_TTL_TX_PORT          PORTE
#define UART_TTL_TX_DDR           DDRE
#define UART_TTL_TX_PIN           PE6 

#define UART_TLL_RX_PORT          PORTE
#define UART_TTL_RX_DDR           DDRE
#define UART_TLL_RX_PORTIN        PINE
#define UART_TTL_RX_PIN           PE7 

// uartsw transmit status and data variables
volatile uint8_t uart_ttl_tx_busy;
volatile uint8_t uart_ttl_tx_data;
volatile uint8_t uart_ttl_tx_bit_num;

// baud rate common to transmit and receive
volatile uint16_t uart_ttl_baudrate_div;

// uartsw receive status and data variables
volatile uint8_t uart_ttl_rx_busy;
volatile uint8_t uart_ttl_rx_data;
volatile uint8_t uart_ttl_rx_bit_num;

/* rx buffer */
unsigned char uart_ttl_rx_data_buffer[UART_TTL_RX_BUFFER_SIZE];
unsigned char uart_ttl_rx_read_ptr;
unsigned char uart_ttl_rx_write_ptr;
unsigned char uart_ttl_rx_num_data;

/* private functions */

/* interrupt handlers */
ISR(TIMER3_CAPT_vect)//RX
{
  uint16_t capture;
  // clear ICP interrupt flag
  ETIFR|=(1<<ICF3);
 
  if(!uart_ttl_rx_busy)
  {
    // disable input capture interrutps
    ETIMSK&=~(1<<TICIE3);
    capture=TCNT3+uart_ttl_baudrate_div + uart_ttl_baudrate_div/2;
    // schedule data bit sampling 1.5 bit periods from now
    OCR3BH=(capture>>8)&0xFF;
    OCR3BL=capture&0xFF;
    // clear OC1B interrupt flag
    ETIFR|=(1<<OCF3B);
    // enable OC1B interrupt
    ETIMSK|=(1<<OCIE3B);
    // set start bit flag
    uart_ttl_rx_busy = 0x01;
    // reset bit counter
    uart_ttl_rx_bit_num = 0;
    // reset data
    uart_ttl_rx_data = 0;
  }
}

ISR(TIMER3_COMPB_vect)//RX
{
  uint16_t capture;

  // clear OC1B interrupt flag
  ETIFR|=(1<<OCF3B);
  // shift data byte to make room for new bit
  uart_ttl_rx_data = uart_ttl_rx_data>>1;

  if(UART_TLL_RX_PORTIN&(1<<UART_TTL_RX_PIN))
    uart_ttl_rx_data |= 0x80;

  // increment bit counter
  uart_ttl_rx_bit_num++;
  // schedule next bit sample
  capture=OCR3B+uart_ttl_baudrate_div;
  OCR3BH=(capture>>8)&0xFF;
  OCR3BL=capture&0xFF;

  // check if we have a full byte
  if(uart_ttl_rx_bit_num >= 8)
  {
    // save data in receive buffer
    uart_ttl_rx_data_buffer[uart_ttl_rx_write_ptr]=uart_ttl_rx_data;
    uart_ttl_rx_write_ptr=(uart_ttl_rx_write_ptr+1)%UART_TTL_RX_BUFFER_SIZE;
    uart_ttl_rx_num_data++;
    // disable OC1B interrupt
    ETIMSK&=~(1<<OCIE3B);
    // enable ICP interrupt
    ETIMSK|=(1<<TICIE3);
    // clear start bit flag
    uart_ttl_rx_busy = 0x00;
  }
}

ISR(TIMER3_COMPA_vect)//TX
{
  uint16_t capture;

  // clear OC1B interrupt flag
  ETIFR|=(1<<OCF3A);
  if(uart_ttl_tx_bit_num)
  {
    if(uart_ttl_tx_bit_num>1)
    {
      // transmit data bits (inverted, LSB first)
      if((uart_ttl_tx_data&0x01))
        UART_TTL_TX_PORT|=(1<<UART_TTL_TX_PIN);
      else
        UART_TTL_TX_PORT&=~(1<<UART_TTL_TX_PIN);
      // shift bits down
      uart_ttl_tx_data = uart_ttl_tx_data>>1;
    }
    else
      UART_TTL_TX_PORT|=(1<<UART_TTL_TX_PIN);
    // schedule the next bit
    capture=OCR3A+uart_ttl_baudrate_div;
    OCR3AH=(capture>>8)&0xFF;
    OCR3AL=capture&0xFF;
    // count down
    uart_ttl_tx_bit_num--;
  }
  else
  {
    // enable OC1A interrupt
    ETIMSK&=~(1<<OCIE3A);
    uart_ttl_tx_busy = 0x00;
  }
}

/* public functions */
void uart_ttl_init(void)
{
  // configure GPIO ports
  UART_TTL_TX_DDR|=(1<<UART_TTL_TX_PIN);// output
  UART_TTL_RX_DDR&=~(1<<UART_TTL_RX_PIN);//input
  // set input to 0
  UART_TLL_RX_PORT&=~(1<<UART_TTL_RX_PIN);
  // initialize baud rate
  uart_ttl_set_baudrate(UART_TTL_DEFAULT_BAUDRATE);
  ram_data[UART_TTL_BAUDRATE_L]=UART_TTL_DEFAULT_BAUDRATE%256;
  ram_data[UART_TTL_BAUDRATE_H]=UART_TTL_DEFAULT_BAUDRATE/256;

  // setup the transmitter
  uart_ttl_tx_busy = 0x00;
  ETIMSK&=~(1<<OCIE3A);

  // setup the receiver
  uart_ttl_rx_busy = 0x00;
  ETIMSK&=~(1<<OCIE3B);
  // trigger on falling edge 
  TCCR3B&=~(1<<ICES3);
  // disable ICP interrupt
  ETIMSK&=~(1<<TICIE3);

  /* initlaize buffer variables */
  uart_ttl_rx_read_ptr=0;
  uart_ttl_rx_write_ptr=0;
  uart_ttl_rx_num_data=0;
}

void uart_ttl_start(void)
{
  ETIMSK&=~(1<<OCIE3A);
  ETIMSK&=~(1<<OCIE3B);
  // clear ICP interrupt flag
  ETIFR|=(1<<ICF3);
  ETIMSK|=(1<<TICIE3);
  ram_data[UART_TTL_CONTROL]|=0x20;
}

void uart_ttl_stop(void)
{
  // disable interrupts
  ETIMSK&=~(1<<OCIE3A);
  ETIMSK&=~(1<<OCIE3B);
  ETIMSK&=~(1<<TICIE3);
  ram_data[UART_TTL_CONTROL]&=0xDF;
}

void uart_ttl_set_baudrate(uint32_t baudrate)
{
  TCCR3B&=0xF8;
  TCCR3B|=0x01;
  // calculate division factor for requested baud rate, and set it
  uart_ttl_baudrate_div = (uint16_t)((F_CPU+(baudrate/2L))/(baudrate*1L));
  ram_data[UART_TTL_BAUDRATE_L]=baudrate%256;
  ram_data[UART_TTL_BAUDRATE_H]=baudrate/256;
}

void uart_ttl_send_byte(uint8_t data)
{
  uint16_t capture;

  // wait until uart is ready
  while(uart_ttl_tx_busy);
  // set busy flag
  uart_ttl_tx_busy = 0x01;
  // save data
  uart_ttl_tx_data = data;
  // set number of bits (+1 for stop bit)
  uart_ttl_tx_bit_num = 9;

  UART_TTL_TX_PORT&=~(1<<UART_TTL_TX_PIN);

  capture=TCNT3+uart_ttl_baudrate_div;
  OCR3AH=(capture>>8)&0xFF;
  OCR3AL=capture&0xFF;
  // clear interrupts
  ETIFR|=(1<<OCF3A);
  // enable OC1A interrupt
  ETIMSK|=(1<<OCIE3A);
}

uint8_t uart_ttl_is_sending(void)
{
  return uart_ttl_tx_busy;
}

uint8_t uart_ttl_receive_byte(uint8_t* data)
{
  // make sure we have a receive buffer
  if(uart_ttl_rx_num_data>0)
  {
    *data=uart_ttl_rx_data_buffer[uart_ttl_rx_read_ptr];
    uart_ttl_rx_read_ptr=(uart_ttl_rx_read_ptr+1)%UART_TTL_RX_BUFFER_SIZE;
    uart_ttl_rx_num_data--;
    if(uart_ttl_rx_num_data==0)
      ram_data[UART_TTL_CONTROL]&=0x7F;
    // make sure we have data
    return 0x01;
  }
  else
    return 0x00;
}

uint8_t uart_ttl_is_data_available(void)
{
  if(uart_ttl_rx_num_data>0)
    return 0x01;
  else
    return 0x00;
}
