#include "exp_board_config.h"
#include <avr/io.h>
#include <avr/interrupt.h>
#include "uart_usb.h"
#include "mem.h"

#ifndef UART_USB_DEFAULT_BAUDRATE
  #define UART_USB_DEFAULT_BAUDRATE           9600
#endif

#ifndef UART_USB_RX_BUFFER_SIZE
  #define UART_USB_RX_BUFFER_SIZE             32
#endif

#define UART_USB_TX_PORT          PORTD
#define UART_USB_TX_DDR           DDRD
#define UART_USB_TX_PIN           PD5 

#define UART_USB_RX_PORT          PORTD
#define UART_USB_RX_DDR           DDRD
#define UART_USB_RX_PORTIN        PIND
#define UART_USB_RX_PIN           PD4 

// uartsw transmit status and data variables
volatile uint8_t uart_usb_tx_busy;
volatile uint8_t uart_usb_tx_data;
volatile uint8_t uart_usb_tx_bit_num;

// baud rate common to transmit and receive
volatile uint16_t uart_usb_baudrate_div;

// uartsw receive status and data variables
volatile uint8_t uart_usb_rx_busy;
volatile uint8_t uart_usb_rx_data;
volatile uint8_t uart_usb_rx_bit_num;

/* rx buffer */
unsigned char uart_usb_rx_data_buffer[UART_USB_RX_BUFFER_SIZE];
unsigned char uart_usb_rx_read_ptr;
unsigned char uart_usb_rx_write_ptr;
unsigned char uart_usb_rx_num_data;

/* private functions */

/* interrupt handlers */
ISR(TIMER1_CAPT_vect)//RX
{
  uint16_t capture;
  // clear ICP interrupt flag
  TIFR|=(1<<ICF1);
  
  if(!uart_usb_rx_busy)
  {
    // disable input capture interrutps
    TIMSK&=~(1<<TICIE1);
    capture=TCNT1+uart_usb_baudrate_div + uart_usb_baudrate_div/2;
    // schedule data bit sampling 1.5 bit periods from now
    OCR1BH=(capture>>8)&0xFF;
    OCR1BL=capture&0xFF;
    // clear OC1B interrupt flag
    TIFR|=(1<<OCF1B);
    // enable OC1B interrupt
    TIMSK|=(1<<OCIE1B);
    // set start bit flag
    uart_usb_rx_busy = 0x01;
    // reset bit counter
    uart_usb_rx_bit_num = 0;
    // reset data
    uart_usb_rx_data = 0;
  }
}

ISR(TIMER1_COMPB_vect)//RX
{
  uint16_t capture;

  // clear OC1B interrupt flag
  TIFR|=(1<<OCF1B);
  // shift data byte to make room for new bit
  uart_usb_rx_data = uart_usb_rx_data>>1;

  if(UART_USB_RX_PORTIN&(1<<UART_USB_RX_PIN))
    uart_usb_rx_data |= 0x80;

  // increment bit counter
  uart_usb_rx_bit_num++;
  // schedule next bit sample
  capture=OCR1B+uart_usb_baudrate_div;
  OCR1BH=(capture>>8)&0xFF;
  OCR1BL=capture&0xFF;

  // check if we have a full byte
  if(uart_usb_rx_bit_num >= 8)
  {
    // save data in receive buffer
    uart_usb_rx_data_buffer[uart_usb_rx_write_ptr]=uart_usb_rx_data;
    uart_usb_rx_write_ptr=(uart_usb_rx_write_ptr+1)%UART_USB_RX_BUFFER_SIZE;
    uart_usb_rx_num_data++;
    // disable OC1B interrupt
    TIMSK&=~(1<<OCIE1B);
    // enable ICP interrupt
    TIMSK|=(1<<TICIE1);
    // clear start bit flag
    uart_usb_rx_busy = 0x00;
  }
}

ISR(TIMER1_COMPA_vect)//TX
{
  uint16_t capture;

  // clear OC1B interrupt flag
  TIFR|=(1<<OCF1A);
  if(uart_usb_tx_bit_num)
  {
    if(uart_usb_tx_bit_num>1)
    {
      // transmit data bits (inverted, LSB first)
      if((uart_usb_tx_data&0x01))
        UART_USB_TX_PORT|=(1<<UART_USB_TX_PIN);
      else
        UART_USB_TX_PORT&=~(1<<UART_USB_TX_PIN);
      // shift bits down
      uart_usb_tx_data = uart_usb_tx_data>>1;
    }
    else
      UART_USB_TX_PORT|=(1<<UART_USB_TX_PIN);
    // schedule the next bit
    capture=OCR1A+uart_usb_baudrate_div;
    OCR1AH=(capture>>8)&0xFF;
    OCR1AL=capture&0xFF;
    // count down
    uart_usb_tx_bit_num--;
  }
  else
  {
    // enable OC1A interrupt
    TIMSK&=~(1<<OCIE1A);
    uart_usb_tx_busy = 0x00;
  }
}

/* public functions */
void uart_usb_init(void)
{
  // configure GPIO ports
  UART_USB_TX_DDR|=(1<<UART_USB_TX_PIN);// output
  UART_USB_RX_DDR&=~(1<<UART_USB_RX_PIN);//input
  // set input to 0
  UART_USB_RX_PORT&=~(1<<UART_USB_RX_PIN);
  // initialize baud rate
  uart_usb_set_baudrate(UART_USB_DEFAULT_BAUDRATE);
  ram_data[UART_TTL_BAUDRATE_L]=UART_USB_DEFAULT_BAUDRATE%256;
  ram_data[UART_TTL_BAUDRATE_H]=UART_USB_DEFAULT_BAUDRATE/256;

  // setup the transmitter
  uart_usb_tx_busy = 0x00;
  TIMSK&=~(1<<OCIE1A);

  // setup the receiver
  uart_usb_rx_busy = 0x00;
  TIMSK&=~(1<<OCIE1B);
  // trigger on falling edge 
  TCCR1B&=~(1<<ICES1);
  // enable ICP interrupt
  TIMSK|=(1<<TICIE1);

  /* initlaize buffer variables */
  uart_usb_rx_read_ptr=0;
  uart_usb_rx_write_ptr=0;
  uart_usb_rx_num_data=0;
}

void uart_usb_start(void)
{
  TIMSK&=~(1<<OCIE1A);
  TIMSK&=~(1<<OCIE1B);
  // clear ICP interrupt flag
  TIFR|=(1<<ICF1);
  TIMSK|=(1<<TICIE1);
  ram_data[UART_USB_CONTROL]|=0x20;
}

void uart_usb_stop(void)
{
  // disable interrupts
  TIMSK&=~(1<<OCIE1A);
  TIMSK&=~(1<<OCIE1B);
  TIMSK&=~(1<<TICIE1);
  ram_data[UART_USB_CONTROL]&=0xDF;
}

void uart_usb_set_baudrate(uint32_t baudrate)
{
  TCCR1B&=0xF8;
  TCCR1B|=0x01;
  // calculate division factor for requested baud rate, and set it
  uart_usb_baudrate_div = (uint16_t)((F_CPU+(baudrate/2L))/(baudrate*1L));
  ram_data[UART_USB_BAUDRATE_L]=baudrate%256;
  ram_data[UART_USB_BAUDRATE_H]=baudrate/256;
}

void uart_usb_send_byte(uint8_t data)
{
  uint16_t capture;

  // wait until uart is ready
  while(uart_usb_tx_busy);
  // set busy flag
  uart_usb_tx_busy = 0x01;
  // save data
  uart_usb_tx_data = data;
  // set number of bits (+1 for stop bit)
  uart_usb_tx_bit_num = 9;

  UART_USB_TX_PORT&=~(1<<UART_USB_TX_PIN);

  capture=TCNT1+uart_usb_baudrate_div;
  OCR1AH=(capture>>8)&0xFF;
  OCR1AL=capture&0xFF;
  // clear interrupts
  TIFR|=(1<<OCF1A);
  // enable OC1A interrupt
  TIMSK|=(1<<OCIE1A);
}

uint8_t uart_usb_is_sending(void)
{
  return uart_usb_tx_busy;
}

uint8_t uart_usb_receive_byte(uint8_t* data)
{
  // make sure we have a receive buffer
  if(uart_usb_rx_num_data>0)
  {
    *data=uart_usb_rx_data_buffer[uart_usb_rx_read_ptr];
    uart_usb_rx_read_ptr=(uart_usb_rx_read_ptr+1)%UART_USB_RX_BUFFER_SIZE;
    uart_usb_rx_num_data--;
    if(uart_usb_rx_num_data==0)
      ram_data[UART_USB_CONTROL]&=0x7F;
    return 0x01;
  }
  else
    return 0x00;
}

uint8_t uart_usb_is_data_available(void)
{
  if(uart_usb_rx_num_data>0)
    return 0x01;
  else
    return 0x00;
}
